﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace UkazkaDekompozice
{
    public interface GraveYardManager
    {
        void burryBodyIntoGrave(Body body, Grave grave);
        void removeBodyFromGrave(Body body, Grave grave);
        Grave findGraveWithBody(Body body);
        IList<Body> getBodiesInGrave(Grave grave);
    }

    public class MemoryGraveYardManager:GraveYardManager
    {
        private BodyManager bodyManager = new MemoryBodyManager();
        private GraveManager graveManager = new MemoryGraveManager();
        private IDictionary<Grave, IList<Body>> bodiesInGraves = new Dictionary<Grave, IList<Body>>();
        public void burryBodyIntoGrave(Body body, Grave grave)
        {
            IList<Body> bodiesInGrave;
            if(!bodiesInGraves.ContainsKey(grave))
            {
                bodiesInGrave = new List<Body>();
                bodiesInGraves.Add(grave, bodiesInGrave);
            }
            bodiesInGrave = bodiesInGraves[grave];
            if (grave.capacity <= bodiesInGrave.Count)
            {
                throw new ArgumentException("Grave is full.");
            }
            bodiesInGrave.Add(body);
        }

        public Grave findGraveWithBody(Body body)
        {
            foreach(Grave grave in bodiesInGraves.Keys)
            {
                if (bodiesInGraves[grave].Contains(body))
                {
                    return grave;
                }
            }
            return null;
        }

        public IList<Body> getBodiesInGrave(Grave grave)
        {
            return bodiesInGraves[grave];
        }

        public void removeBodyFromGrave(Body body, Grave grave)
        {
            Grave graveToRemove = findGraveWithBody(body);
            IList<Body> bodies = bodiesInGraves[graveToRemove];

            if (!bodies.Remove(body))
            {
                throw new ArgumentException("No such body in grave.");
            }
        }
    }

    public class Body
    {
        public long id { get; set; }
        public string name { get; set; }
        public DateTime birthDate { get; set; }
        public DateTime deathDate { get; set; } 
        public bool isVampire { get; set; }

        public override bool Equals(object obj)
        {
            if(!(obj is Body))
            {
                return false;
            }
            Body body = obj as Body;
            return id == body.id;
        }

        public override int GetHashCode()
        {
            return id.GetHashCode();
        }

        public override string ToString()
        {
            return name + ", born: " + birthDate + ", died: " + deathDate + ", upir: " + (isVampire?"ano":"ne");
        }
    }

    public interface BodyManager
    {
        void createBody(Body body);
        void updateBody(Body body);
        void removeBody(Body body);
        Body findBodyWithId(long id);
        IList<Body> findBodiesByName(string name);
        IList<Body> findAllBodies();
    }

    public class MemoryBodyManager:BodyManager
    {
        private List<Body> bodies = new List<Body>();

        public void createBody(Body body)
        {
            bodies.Add(body);
        }

        public IList<Body> findAllBodies()
        {
            return bodies.AsReadOnly();
        }

        public IList<Body> findBodiesByName(string name)
        {
            IList<Body> bodies = new List<Body>();
            foreach(Body body in bodies)
            {
                if(body.name == name)
                {
                    bodies.Add(body);
                }
            }
            return bodies;
        }

        public Body findBodyWithId(long id)
        {
            foreach(Body body in bodies)
            {
                if(body.id == id)
                {
                    return body;
                }
            }
            return null;
        }

        public void removeBody(Body body)
        {
            if (!bodies.Contains(body))
            {
                throw new ArgumentException("No such body.");
            }
            bodies.Remove(body);
        }

        public void updateBody(Body body)
        {
            foreach(Body b in bodies)
            {
                if(b == body)
                {
                    b.isVampire = body.isVampire;
                    b.name = body.name;
                    b.birthDate = body.birthDate;
                    b.deathDate = body.deathDate;
                    return;
                }
            }
            throw new ArgumentException("No such body.");
        }

        
    }

    public class Grave
    {
        public long id { get; set; }
        public int row { get; set; }
        public int column { get; set; }
        public int capacity { get; set; }

        public override bool Equals(object obj)
        {
            if (!(obj is Grave)) return false;
            Grave grave = obj as Grave;
            return id == grave.id;
        }

        public override int GetHashCode()
        {
            return id.GetHashCode();
        }

        public override string ToString()
        {
            return "Grave: " + id + ": row=" + row + ", column=" + column + ", capacity=" + capacity;
        }
    }

    public interface GraveManager
    {
        void createGrave(Grave grave);
        void updateGrave(Grave grave);
        void removeGrave(Grave grave);
        Grave findGraveById(int id);
        IList<Grave> findAllGraves();

    }

    public class MemoryGraveManager : GraveManager
    {

        private List<Grave> graves = new List<Grave>();

        public void createGrave(Grave grave)
        {
            if (grave == null)
            {
                throw new ArgumentNullException("grave");
            }
            graves.Add(grave);
        }

        public IList<Grave> findAllGraves()
        {
            return graves.AsReadOnly();
        }

        public Grave findGraveById(int id)
        {
            foreach(Grave grave in graves)
            {
                if(grave.id == id)
                {
                    return grave;
                }
            }
            return null;
        }

        public void removeGrave(Grave grave)
        {
            if(!graves.Remove(grave))
            {
                throw new ArgumentException("No such grave.");
            }
        }

        public void updateGrave(Grave grave)
        {
            Grave updated = graves.Find(x => x.id == grave.id);
            if(updated == null)
            {
                throw new ArgumentException("No such grave.");
            }
            graves.Remove(updated);
            graves.Add(grave);
        }
    }

    public class Program
    {
        static void Main(string[] args)
        {
            GraveYardManager manager = new MemoryGraveYardManager();
            Body bodyOne = new Body(),
                bodyTwo = new Body();
            bodyOne.name = "Vlad Tepesz";
            bodyOne.birthDate = new DateTime(1700, 01, 01);
            bodyOne.deathDate = new DateTime(1970, 12, 31);
            bodyOne.id = 1;
            bodyOne.isVampire = true;
            Grave grave = new Grave();
            grave.id = 1;
            grave.column = 1;
            grave.row = 1;
            grave.capacity = 2;
            try
            {
                manager.burryBodyIntoGrave(bodyOne, grave);
            }
            catch(Exception ex)
            {
                Console.Error.WriteLine("Cannot burry first body: " + ex);
            }

            bodyTwo.name = "Obet Lucy";
            bodyTwo.id = 2;
            bodyTwo.isVampire = false;
            bodyTwo.birthDate = new DateTime(1900, 02, 02);
            bodyTwo.deathDate = new DateTime(1932, 01, 01);
            try
            {
                manager.burryBodyIntoGrave(bodyTwo, grave);
            }
            catch(Exception ex)
            {
                Console.Error.WriteLine("Cannot burry second body:" + ex);
            }

            Console.WriteLine(manager.findGraveWithBody(bodyTwo));
            foreach(Body body in manager.getBodiesInGrave(grave))
            {
                Console.WriteLine(body);
            }

            try
            {
                manager.burryBodyIntoGrave(bodyTwo, grave);
                Console.Error.WriteLine("Burrying third body into grave should fail.");
            }
            catch(Exception ex)
            {
                Console.WriteLine("Unable to burry third body into grave with capacity 2.");
            }
            Console.ReadLine();

        }
    }
}
